# coding=utf-8

import time
import threading
import serial
import serial.serialutil


from pymycobot.mercury_api import MercuryCommandGenerator
from pymycobot.error import calibration_parameters


class Mercury(MercuryCommandGenerator):
    def __init__(self, port, baudrate="115200", timeout=0.1, debug=False):
        """
        Args:
            port     : port string
            baudrate : baud rate string, default '115200'
            timeout  : default 0.1
            debug    : whether show debug info
        """
        super(Mercury, self).__init__(debug)
        self.calibration_parameters = calibration_parameters

        self._serial_port = serial.Serial()
        self._serial_port.port = port
        self._serial_port.baudrate = baudrate
        self._serial_port.timeout = timeout
        self._serial_port.rts = False
        self._serial_port.open()
        self.lock = threading.Lock()
        self.lock_out = threading.Lock()
        self.has_reply_command = []
        self.is_stop = False
        self.read_threading = threading.Thread(target=self.read_thread)
        self.read_threading.daemon = True
        self.read_threading.start()
        try:
            self._serial_port.write(b"\xfa")
        except serial.serialutil.SerialException as e:
            self._serial_port.close()
            time.sleep(0.5)
            self._serial_port.open()
        self.get_limit_switch()

    def open(self):
        self._serial_port.open()
        
    def close(self):
        self._serial_port.close()



# coding=utf-8
import functools
from pymycobot.robot_info import RobotLimit
# In order to adapt to the atom side, ID of 0-5 or 1-6 are allowed.
# In order to support end control, ID 7 is allowed.
MIN_ID = 0
MAX_ID = 7

# In fact, most joints cannot reach plus or minus 180 angles.
# There may be a value greater than 180 when reading the angle,
# and the maximum and minimum values are expanded for compatibility.
MIN_ANGLE = -190.0
MAX_ANGLE = 190.0


class MyCobotDataException(Exception):
    pass

class MercuryDataException(Exception):
    pass

class MyAgvDataException(Exception):
    pass

class MechArmDataException(Exception):
    pass

class MyArmDataException(Exception):
    pass

class UltraArmDataException(Exception):
    pass

class MyPalletizerDataException(Exception):
    pass

class MyBuddyDataException(Exception):
    pass

class MercuryRobotException(Exception):
    pass

def check_boolean(b):
    if b != 0 and b != 1:
        raise MyCobotDataException("This parameter needs to be 0 or 1")
    
def check_rgb_value(value, exception_class, class_name):
    rgb_str = ["r", "g", "b"]
    for i, v in enumerate(value):
        if not (0 <= v <= 255):
            raise exception_class(
                "The RGB value for {} needs to be 0 ~ 255, but received the {} is {}".format(class_name, rgb_str[i], v)
            )

def check_value_type(parameter, value_type, exception_class, _type):
    if value_type is not _type:
        raise exception_class("The acceptable parameter {} should be an {}, but the received {}".format(parameter, _type, value_type))
    
def check_coords(value, robot_limit, class_name, exception_class):
    if len(value) != 6:
        raise exception_class("The length of `coords` must be 6.")
    for idx, coord in enumerate(value):
        if not robot_limit[class_name]["coords_min"][idx] <= coord <= robot_limit[class_name]["coords_max"][idx]:
            raise exception_class(
                "Has invalid coord value, error on index {0}. received {3} .but angle should be {1} ~ {2}.".format(
                    idx, robot_limit[class_name]["coords_min"][idx], robot_limit[class_name]["coords_max"][idx], coord
                )
            )
            
def check_0_or_1(parameter, value, range_data, value_type, exception_class, _type):
    check_value_type(parameter, value_type, exception_class, _type)
    if value not in range_data:
        error = "The data supported by parameter {} is ".format(parameter)
        lens = len(range_data)
        for idx in range(lens):
            error += str(range_data[idx])
            if idx != lens - 1:
                error += " or "
        error += ", but the received value is {}".format(value)
        raise exception_class(error)
    
def check_id(value, id_list, exception_class):
    raise exception_class(
        "The joint_id not right, should be in {0}, but received {1}.".format(
        id_list, value
        )
    )
    
def public_check(parameter_list, kwargs, robot_limit, class_name, exception_class):
    for parameter in parameter_list[1:]:
        value = kwargs.get(parameter, None)
        value_type = type(value)
        if parameter == 'id' and value not in robot_limit[class_name][parameter]:
            check_id(value, robot_limit[class_name][parameter], exception_class)
        elif parameter == 'rgb':
            check_rgb_value(value, exception_class, class_name)
        elif parameter == 'value':
            check_value_type(parameter, value_type, exception_class, int)
            if value < 0 or value > 4096:
                raise exception_class("The range of {} is 0 ~ 4096, but the received value is {}".format(parameter, value))
        elif parameter == 'pin_mode':
            check_value_type(parameter, value_type, exception_class, int)
            if value not in [0,1,2]:
                raise exception_class("The data supported by parameter {} is 0 or 1 or 2, but the received value is {}".format(parameter, value))
        elif parameter == 'pin_signal':
            check_0_or_1(parameter, value, [0, 1], value_type, exception_class, int)
        elif parameter == 'speed':
            check_value_type(parameter, value_type, exception_class, int)
            if not 1 <= value <= 100: 
                raise exception_class(
                    "speed value not right, should be 1 ~ 100, the received speed is %s"
                    % value
                )
        elif parameter == 'flag':
            check_0_or_1(parameter, value, [0, 1, 254], value_type, exception_class, int)
        elif parameter == 'gripper_type':
            check_0_or_1(parameter, value, [1, 3, 4], value_type, exception_class, int)
        elif parameter == '_type_1':
            check_0_or_1(parameter, value, [1, 2, 3, 4], value_type, exception_class, int)
            # if value not in [0, 1, 10]:
            #     raise exception_class("The data supported by parameter {} is 0 or 1 or 10, but the received value is {}".format(parameter, value))
        elif parameter == 'gripper_value':
            check_value_type(parameter, value_type, exception_class, int)
            if value < 0 or value > 100:
                raise exception_class("The range of {} is 0 ~ 100, but the received value is {}".format(parameter, value))
        elif parameter in ['account', 'password']:
            check_value_type(parameter, value_type, exception_class, str)
        # TODO 280/320共用MyCobot，无法进行数据限位
        # elif parameter == 'coords':
        #     check_coords(value, robot_limit, class_name, exception_class)
        elif parameter in ['rftype', 'move_type', 'end', 'is_linear', 'status', 'mode', 'direction', 'state', 'deceleration']:
            check_0_or_1(parameter, value, [0, 1], value_type, exception_class, int)
        elif parameter == 'acceleration':
            check_value_type(parameter, value_type, exception_class, int)
            if not 1 <= value <= 100:
                raise exception_class(
                    "{} value not right, should be 1 ~ 100, the received is {}".format(parameter, value)
                )
        elif parameter == 'angle':
            joint_id = kwargs.get('id', None)
            index = robot_limit[class_name]['joint_id'][joint_id-1] - 1
            if value < robot_limit[class_name]["angles_min"][index] or value > robot_limit[class_name]["angles_max"][index]:
                raise exception_class(
                    "angle value not right, should be {0} ~ {1}, but received {2}".format(
                        robot_limit[class_name]["angles_min"][index], robot_limit[class_name]["angles_max"][index], value
                    )
                )
        elif parameter == 'encoders':
            if  "MyCobot" in class_name or "MechArm" in class_name:
                if len(value) != 6:
                    raise exception_class("The length of `encoders` must be 6.")
            elif "MyPalletizer" in class_name:
                if len(value) != 4:
                    raise exception_class("The length of `encoders` must be 4.")
            elif "MyPalletizer" in class_name:
                if len(value) != 7:
                    raise exception_class("The length of `encoders` must be 7.")
            for data in value:
                data_type = type(data)
                check_value_type(parameter, data_type, exception_class, int)
                if data < 0 or data > 4096:
                    raise exception_class("The range of encoder is 0 ~ 4096, but the received value is {}".format(data))
        elif parameter == 'speeds':
            if  "MyCobot" in class_name or "MechArm" in class_name:
                if len(value) not in [6, 7]:
                    raise exception_class("The length of `speeds` must be 6. but the received value is {}".format(value))
            elif "MyPalletizer" in class_name:
                if len(value) != 4:
                    raise exception_class("The length of `speeds` must be 4.")
            elif "MyArm" in class_name:
                if len(value) != 7:
                    raise exception_class("The length of `speeds` must be 7.")
            for data in value:
                data_type = type(data)
                check_value_type(parameter, data_type, exception_class, int)
                if data < 0 or data > 3400:
                    raise exception_class("The range of speed is 0 ~ 3400, but the received value is {}".format(data))
        elif parameter in ['servo_id_pdi', 'encode_id']:
            check_value_type(parameter, value_type, exception_class, int)
            # if  "MyCobot" in class_name or "MechArm" in class_name:
            #     if value < 1 or value > 6:
            #             raise exception_class("The range of joint_id is 1 ~ 6, but the received is {}".format(value))
            if "MyPalletizer" in class_name:
                if value < 1 or value > 4:
                        raise exception_class("The range of joint_id is 1 ~ 4, but the received is {}".format(value))
            elif "MyArm" in class_name or "MyCobot" in class_name or "MechArm" in class_name:
                if value < 1 or value > 7:
                        raise exception_class("The range of joint_id is 1 ~ 7, but the received is {}".format(value))
        elif parameter == "torque":
            torque_min = 150
            torque_max = 980
            if value < torque_min or value > torque_max:
                raise exception_class("The range of torque is {} ~ {}, but the received is {}".format(torque_min, torque_max, value))
        elif parameter == "current":
            current_min = 1
            current_max = 500
            if value < current_min or value > current_max:
                raise exception_class("The range of current is {} ~ {}, but the received is {}".format(current_min, current_max, value))
        elif parameter == 'end_direction':
            check_0_or_1(parameter, value, [1, 2, 3], value_type, exception_class, int)
        elif parameter == "pin_no":
            if  "Mercury" in class_name:
                check_0_or_1(parameter, value, [1, 2, 3, 4, 5, 6], value_type, exception_class, int)
def calibration_parameters(**kwargs):
    # with open(os.path.dirname(os.path.abspath(__file__))+"/robot_limit.json") as f:
    robot_limit = RobotLimit.robot_limit
    parameter_list = list(kwargs.keys())
    class_name =  kwargs.get("class_name", None)
    if class_name in ["Mercury", "MercurySocket"]:
        for parameter in parameter_list[1:]:
            value = kwargs.get(parameter, None)
            if parameter == 'joint_id':
                if value not in robot_limit[class_name][parameter]:
                    check_id(value, robot_limit[class_name][parameter], MercuryDataException)
            elif parameter == 'degree':
                joint_id = kwargs.get('joint_id', None)
                if joint_id in [11,12,13]:
                    index = robot_limit[class_name]['joint_id'][joint_id-4] - 4
                else:
                    index = robot_limit[class_name]['joint_id'][joint_id-1] - 1
                if value < robot_limit[class_name]["angles_min"][index] or value > robot_limit[class_name]["angles_max"][index]:
                    raise MercuryDataException(
                        "angle value not right, should be {0} ~ {1}, but received {2}".format(
                            robot_limit[class_name]["angles_min"][index], robot_limit[class_name]["angles_max"][index], value
                        )
                    )
            elif parameter == 'angles':
                if len(value) not in [7, 10]:
                    raise MercuryDataException("The length of `angles` must be 7 or 10.")
                for idx, angle in enumerate(value):
                    joint_id = idx+1
                    angle_min = robot_limit[class_name]["angles_min"][idx]
                    angle_max = robot_limit[class_name]["angles_max"][idx]
                    if angle < angle_min or angle > angle_max:
                        raise MercuryDataException(
                            "Joint {} angle value of {} exceeds the limit, with a limit range of {} ~ {}.".format(joint_id, angle, angle_min, angle_max)
                        )

            elif parameter == 'coord':
                
                index = kwargs.get('coord_id', None) - 1
                if value < robot_limit[class_name]["coords_min"][index] or value > robot_limit[class_name]["coords_max"][index]:
                    raise MercuryDataException(
                        "The coord value of {} exceeds the limit, and the limit range is {} ~ {}".format(
                            value, robot_limit[class_name]["coords_min"][index], robot_limit[class_name]["coords_max"][index]
                        )
                    )
            elif parameter == 'coords':
                check_coords(value, robot_limit, class_name, MercuryDataException)

            elif parameter == 'speed' and not 1 <= value <= 100:
                raise MercuryDataException(
                    "speed value not right, should be 1 ~ 100, the error speed is %s"
                    % value
                )

            elif parameter == 'rgb':
                check_rgb_value(value, MercuryDataException, class_name)
            # if direction is not None:
            elif parameter in ['direction', 'flag', 'value']:
                if value not in [0, 1]:
                    raise MercuryDataException("{} only supports 0 or 1, but received {}".format(parameter, value))
                
            elif parameter == 'coord_id':
                if value < 1 or value > 6:
                    raise MercuryDataException("coord_id only supports 1 ~ 6, but received {}".format(value))
                
            elif parameter == 'solution_angle':
                if value > 90 or value < -90:
                    raise MercuryDataException("The angle range is -90° ~ 90°, but received {}".format(value))
            elif parameter == 'address':
                if value < 32 or value > 34:
                    raise MercuryDataException("The angle address is 32 ~ 34, but received {}".format(value))
            elif parameter == 'value':
                if value < 1 or value > 32000:
                    raise MercuryDataException("The angle value is 1 ~ 32000, but received {}".format(value))
            elif parameter == "servo_restore":
                if value not in [1,2,3,4,5,6,7,13,254]:
                    raise MercuryDataException("The joint_id should be in [1,2,3,4,5,6,7,13,254], but received {}".format(value))
            elif parameter == "data_len":
                if value < 1 or value > 45:
                    raise MercuryDataException("The parameter data_len data range only supports 1 ~ 45, but received {}".format(value))
            elif parameter == "max_time":
                if value < 0:
                    raise MercuryDataException("The parameter max_time must be greater than or equal to 0, but received {}".format(value))
            elif parameter == "limit_mode":
                if value not in [1,2]:
                    raise MercuryDataException("The parameter {} only supports 1, 2, but received {}".format(parameter, value))
            elif parameter == "_type":
                if value not in [1,2,3,4]:
                    raise MercuryDataException("The parameter {} only supports 1, 2, 3, 4, but received {}".format(parameter, value))
            elif parameter == "axis":
                if value not in [1,2,3]:
                    raise MercuryDataException("The parameter {} only supports 1, 2, 3, but received {}".format(parameter, value))
            elif parameter == "threshold_value":
                if value < 50 or value > 250:
                    raise MercuryDataException("The parameter {} only supports 50 ~ 250, but received {}".format(parameter, value))
            elif parameter == "comp_value":
                if value < 0 or value > 250:
                    raise MercuryDataException("The parameter {} only supports 0 ~ 250, but received {}".format(parameter, value))
            elif parameter == "shoot_value":
                if value < -300 or value > 300:
                    raise MercuryDataException("The parameter {} only supports -300 ~ 300, but received {}".format(parameter, value))
            elif parameter == "head_id":
                if value not in [11,12]:
                    raise MercuryDataException("The parameter {} only supports 11, 12, but received {}".format(parameter, value))
            elif parameter == "err_angle":
                if value < 0 or value > 5:
                    raise MercuryDataException("The parameter {} only supports 0 ~ 5, but received {}".format(parameter, value))
            elif parameter == "r":
                if value < 0 or value > 655.5:
                    raise MercuryDataException("The parameter {} only supports 0 ~ 655.5, but received {}".format(parameter, value))
            elif parameter == "rank":
                if value not in [0,1,2]:
                    raise MercuryDataException("The parameter {} only supports 0 or 1 or 2, but received {}".format(parameter, value))
            elif parameter == "move_type":
                if value not in [0,2]:
                    raise MercuryDataException("The parameter {} only supports 0 or 2, but received {}".format(parameter, value))
            elif parameter == "trajectory":
                if value not in [0,1,2,3,4]:
                    raise MercuryDataException("The parameter {} only supports [0,1,2,3,4], but received {}".format(parameter, value))
            else:
                public_check(parameter_list, kwargs, robot_limit, class_name, MercuryDataException)
    elif class_name == "MyAgv":
        for parameter in parameter_list[1:]:
            value = kwargs.get(parameter, None)
            if parameter == 'rgb':
                check_rgb_value(value, MyAgvDataException, class_name)
            elif parameter == 'led_mode':
                if value not in [1, 2]:
                    raise MyAgvDataException("led mode only supports 1 or 2, but received is {}".format(value))
            elif parameter == 'direction_1':
                if value < 0 or value > 255:
                    raise MyAgvDataException("The range of direction_1 is 0 ~ 255, but the received value is {}".format(value))
            elif parameter == 'direction_2':
                if value < 0 or value > 255:
                    raise MyAgvDataException("The range of direction_2 is 0 ~ 255, but the received value is {}".format(value))
            elif parameter == 'direction_3':
                if value < 0 or value > 255:
                    raise MyAgvDataException("The range of direction_3 is 0 ~ 255, but the received value is {}".format(value))
            elif parameter == 'data':
                if value < 1 or value > 127:
                    raise MyAgvDataException("The range of {} is 1 ~ 127, but the received value is {}".format(parameter, value))
    
    elif class_name in ["MyCobot", "MyCobotSocket"]:
        public_check(parameter_list, kwargs, robot_limit, class_name, MyCobotDataException)
    elif class_name in ["MechArm", "MechArmSocket"]:
        public_check(parameter_list, kwargs, robot_limit, class_name, MechArmDataException)
    elif class_name in ["MyArm", "MyArmSocket"]:
        public_check(parameter_list, kwargs, robot_limit, class_name, MyArmDataException)
        

def restrict_serial_port(func):
    """
    装饰器，用于限制特定串口号的函数调用。
    """
    @functools.wraps(func)
    def wrapper(instance, *args, **kwargs):
        if instance._serial_port.port not in ["/dev/left_arm", "/dev/right_arm"]:
            raise MercuryRobotException(f"The {func.__name__} function cannot be called. This function is only applicable to the Mercury dual-arm robot.")
        return func(*args, **kwargs)
    return wrapper


# coding=utf-8

from __future__ import division

import sys
import logging
import threading
from pymycobot.common import ProtocolCode, write, read, DataProcessor
from pymycobot.error import calibration_parameters
from pymycobot.log import setup_logging
import serial


class MyArmAPI(DataProcessor):
    """

    """

    def __init__(self, port, baudrate="115200", timeout=0.1, debug=False):
        """
        Args:
            port     : port string
            baudrate : baud rate string, default '115200'
            timeout  : default 0.1
            debug    : whether show debug info
        """
        super(MyArmAPI, self).__init__()
        self.calibration_parameters = calibration_parameters
        self._serial_port = serial.Serial()
        self._serial_port.port = port
        self._serial_port.baudrate = baudrate
        self._serial_port.timeout = timeout
        self._serial_port.rts = False
        self._serial_port.open()
        self.lock = threading.Lock()
        self._version = sys.version_info[:2][0]
        self.debug = debug
        setup_logging(self.debug)
        self.log = logging.getLogger(__name__)

    _write = write
    _read = read

    def _mesg(self, genre, *args, **kwargs):
        """
        Args:
            genre: command type (Command)
            *args: other data.
                   It is converted to octal by default.
                   If the data needs to be encapsulated into hexadecimal,
                   the array is used to include them. (Data cannot be nested)
            **kwargs: support `has_reply`
                has_reply: Whether there is a return value to accept.
        """

        real_command, has_reply = super(MyArmAPI, self)._mesg(genre, *args, **kwargs)
        command = self._flatten(real_command)
        with self.lock:
            self._write(command)

            if not has_reply:
                return None

            timeout = kwargs.get('timeout', None)
            data = self._read(genre, command=command, timeout=timeout)
            res = self._process_received(data, genre, arm=8)

            if len(res) == 0:
                return None

            return_single_genres = [
                ProtocolCode.GET_SERVO_MOTOR_COUNTER_CLOCKWISE,
                ProtocolCode.IS_MOVING,
                ProtocolCode.IS_POWER_ON,
                ProtocolCode.GET_ROBOT_MODIFIED_VERSION,
                ProtocolCode.GET_ROBOT_ATOM_MODIFIED_VERSION,
                ProtocolCode.GET_ROBOT_ERROR_CHECK_STATE,
                ProtocolCode.POWER_OFF,
                ProtocolCode.GET_RECV_QUEUE_SIZE,
                ProtocolCode.GET_RECV_QUEUE_LENGTH,
                ProtocolCode.COBOTX_GET_ANGLE,
                ProtocolCode.GET_ENCODER,
                ProtocolCode.SERVO_RESTORE,
                ProtocolCode.GET_ERROR_DETECT_MODE,
                ProtocolCode.GET_SERVO_MOTOR_CLOCKWISE,
                ProtocolCode.GET_SERVO_MOTOR_CONFIG,
                ProtocolCode.POWER_ON,
                ProtocolCode.GET_MASTER_PIN_STATUS,
                ProtocolCode.GET_ATOM_PIN_STATUS
            ]

            if genre in return_single_genres:
                return self._process_single(res)
            elif genre in [ProtocolCode.GET_ANGLES, ProtocolCode.GET_JOINT_MAX_ANGLE, ProtocolCode.GET_JOINT_MIN_ANGLE]:
                return [self._int2angle(angle) for angle in res]
            elif genre in [ProtocolCode.GET_SERVO_VOLTAGES]:
                return [self._int2coord(angle) for angle in res]
            elif genre == ProtocolCode.GET_ROBOT_ERROR_STATUS:
                result = []
                for item in res:
                    if item == 0:
                        result.append(0)
                        continue

                    reverse_bins = reversed(bin(item)[2:])
                    rank = [i for i, e in enumerate(reverse_bins) if e != '0']
                    result.append(rank)
                return result
            else:
                return res

    def get_robot_modified_version(self):
        """Get the bot correction version number

        Returns:
                version (int): the bot correction version number
        """
        return self._mesg(ProtocolCode.GET_ROBOT_MODIFIED_VERSION, has_reply=True)

    def get_robot_firmware_version(self):
        """Obtaining the Robot Firmware Version (Major and Minor Versions)

        Returns:
                version (int): the robot firmware
        """
        return self._mesg(ProtocolCode.GET_ROBOT_FIRMWARE_VERSION, has_reply=True)

    def get_robot_tool_modified_version(self):
        """Get the remediation version of the bot tool"""
        return self._mesg(ProtocolCode.GET_ROBOT_ATOM_MODIFIED_VERSION, has_reply=True)

    def get_robot_tool_firmware_version(self):
        """Get the Robot Tool Firmware Version (End Atom)"""
        return self._mesg(ProtocolCode.GET_ROBOT_TOOL_FIRMWARE_VERSION, has_reply=True)

    # def get_robot_serial_number(self):
    #     """Get the bot number"""
    #     return self._mesg(ProtocolCode.GET_ROBOT_SERIAL_NUMBER, has_reply=True)

    def set_robot_err_check_state(self, status):
        """Set Error Detection Status You can turn off error detection, but do not turn it off unless necessary

        Args:
            status (int): 1 open; o close
        """
        self._mesg(ProtocolCode.SET_ROBOT_ERROR_CHECK_STATE, status)

    def get_robot_err_check_state(self):
        """Read error detection status"""
        return self._mesg(ProtocolCode.GET_ROBOT_ERROR_CHECK_STATE, has_reply=True)

    def get_robot_error_status(self):
        """Get the bot error status

        Returns:
            No error: [0,0,0,0,0,0,0,0]
            assuming that error 1 and 3 are reported in section 1, it should return: [[1,3],0,0,0,0,0,0,0]
        """
        return self._mesg(ProtocolCode.GET_ROBOT_ERROR_STATUS, has_reply=True, timeout=15)

    def get_robot_power_status(self):
        """Get the robot power status
        Returns:
            power_status (int): 0: power off, 1: power on
        """
        return self._mesg(ProtocolCode.IS_POWER_ON, has_reply=True)

    def set_robot_power_on(self):
        """Set the robot to power on

        Returns: (int) 1
        """
        return self._mesg(ProtocolCode.POWER_ON, has_reply=True)

    def set_robot_power_off(self):
        """Set the robot to power off

        Returns: (int) 1
        """
        return self._mesg(ProtocolCode.POWER_OFF, has_reply=True)

    def clear_robot_err(self):
        """Clear the robot abnormality Ignore the error joint and continue to move"""
        self._mesg(ProtocolCode.CLEAR_ROBOT_ERROR)

    def get_recv_queue_max_len(self):
        """The total length of the read command queue, the default length is 100"""
        return self._mesg(ProtocolCode.GET_RECV_QUEUE_SIZE, has_reply=True)

    def set_recv_queue_max_len(self, max_len):
        """Set the total length of the receiving command queue"""
        self._mesg(ProtocolCode.SET_RECV_QUEUE_SIZE, max_len, has_reply=True)

    def clear_recv_queue(self):
        """Clear the queue for receiving commands"""
        self._mesg(ProtocolCode.CLEAR_RECV_QUEUE)

    def get_recv_queue_len(self):
        """The current length of the read receive queue"""
        return self._mesg(ProtocolCode.GET_RECV_QUEUE_LENGTH, has_reply=True)

    def get_joint_angle(self, joint_id):
        """
        Gets the current angle of the specified joint
        Args:
            joint_id (int): 0 - 254

        Returns:

        """
        return self._mesg(ProtocolCode.COBOTX_GET_ANGLE, joint_id, has_reply=True)

    def get_joints_angle(self):
        """Gets the current angle of all joints

        Returns:
            angles list(int): 0 - 254
        """
        return self._mesg(ProtocolCode.GET_ANGLES, has_reply=True)

    def get_joints_max(self):
        """Read the maximum angle of all joints"""
        max_angles = self._mesg(ProtocolCode.GET_JOINT_MAX_ANGLE, has_reply=True)
        return max_angles

    def get_joints_min(self):
        """Read the minimum angle of all joints"""
        min_angles = self._mesg(ProtocolCode.GET_JOINT_MIN_ANGLE, has_reply=True)
        return list(map(lambda n: n / 100, min_angles))

    def set_servo_calibrate(self, servo_id):
        """Sets the zero position of the specified servo motor

        Args:
            servo_id (int): 0 - 254
        """

        self._mesg(ProtocolCode.SET_SERVO_CALIBRATION, servo_id)

    def get_servo_encoder(self, servo_id):
        """Gets the current encoder potential value for the specified servo motor

        Args:
            servo_id (int): 0 - 254

        Returns:
            encoder (int): 0-4095
        """
        return self._mesg(ProtocolCode.GET_ENCODER, servo_id, has_reply=True)

    def is_all_servos_enabled(self):
        """
        Get the connection status of multiple servo motors

        Returns:
            status: list[int*8] 0: The connection failed 1: The connection is successful
        """
        return self._mesg(ProtocolCode.IS_ALL_SERVO_ENABLE, has_reply=True)

    def get_servos_temp(self):
        """Obtain the temperature of multiple servo motors"""
        return self._mesg(ProtocolCode.GET_SERVO_TEMPS, has_reply=True)

    def get_servos_voltage(self):
        """Get the voltage of multiple servo motors"""
        return self._mesg(ProtocolCode.GET_SERVO_VOLTAGES, has_reply=True)

    def get_servos_current(self):
        """Obtain the current of multiple servo motors"""
        return self._mesg(ProtocolCode.GET_SERVO_CURRENTS, has_reply=True)

    def get_servos_status(self):
        """Get all the statuses of multiple servo motors"""
        return self._mesg(ProtocolCode.GET_SERVO_STATUS, has_reply=True)

    def get_servos_protect_current(self):
        """Obtain multiple servo motor protection currents"""
        return self._mesg(ProtocolCode.GET_SERVO_LASTPDI, has_reply=True)

    def set_servo_enabled(self, joint_id, state):
        """Set the servo motor torque switch

        Args:
            joint_id (int): 0-254 254-all
            state: 0/1
                1-focus
                0-release
        """
        self._mesg(ProtocolCode.RELEASE_ALL_SERVOS, joint_id, state)

    def get_servos_encoder(self):
        """Obtain the current encoder potential values for multiple servo motors"""
        return self._mesg(ProtocolCode.GET_ENCODERS, has_reply=True)

    def get_servos_speed(self):
        """Gets the current movement speed of multiple servo motors"""
        return self._mesg(ProtocolCode.GET_SERVO_SPEED, has_reply=True)

    def set_servo_p(self, servo_id, data):
        """Sets the proportionality factor of the position loop P of the specified servo motor

        Args:
            servo_id (int): 0-254
            data (int): 0-254

        """
        self._mesg(ProtocolCode.MERCURY_DRAG_TECH_SAVE, servo_id, data)

    def get_servo_p(self, servo_id):
        """Reads the position loop P scale factor of the specified servo motor

        Args:
            servo_id (int): 0-254
        """
        return self._mesg(ProtocolCode.SERVO_RESTORE, servo_id, has_reply=True)

    def set_servo_d(self, servo_id, data):
        """Sets the proportional factor for the position ring D of the specified servo motor

        Args:
            servo_id (int): 0-254
            data (int): 0-254
        """
        self._mesg(ProtocolCode.MERCURY_DRAG_TECH_EXECUTE, servo_id, data)

    def get_servo_d(self, servo_id):
        """Reads the position ring D scale factor for the specified servo motor

        Args:
            servo_id (int): 0-254
        """
        return self._mesg(ProtocolCode.SET_ERROR_DETECT_MODE, servo_id, has_reply=True)

    def set_servo_i(self, servo_id, data):
        """Set the proportional factor of the position ring I of the specified servo motor

        Args:
            servo_id (int): 0 - 254
            data (int): 0 - 254
        """
        self._mesg(ProtocolCode.MERCURY_DRAG_TECH_PAUSE, servo_id, data)

    def get_servo_i(self, servo_id):
        """Reads the position loop I scale factor of the specified servo motor"""
        return self._mesg(ProtocolCode.GET_ERROR_DETECT_MODE, servo_id, has_reply=True)

    def set_servo_cw(self, servo_id, data):
        """Sets the clockwise insensitivity zone of the encoder for the specified servo motor

        Args:
            servo_id (int): 0 - 254
            data (int): 0 - 32
        """
        self._mesg(ProtocolCode.SET_SERVO_MOTOR_CLOCKWISE, servo_id, data)

    def get_servo_cw(self, servo_id):
        """Reads the clockwise insensitive area of the encoder for the specified servo motor

        Args:
            servo_id (int): 0 - 254
        """
        return self._mesg(ProtocolCode.GET_SERVO_MOTOR_CLOCKWISE, servo_id, has_reply=True)

    def set_servo_cww(self, servo_id, data):
        """Sets the counterclockwise insensitive zone of the encoder for the specified servo motor

        Args:
            servo_id (int): 0 - 254
            data (int): 0 - 32
        """
        return self._mesg(ProtocolCode.SET_SERVO_MOTOR_COUNTER_CLOCKWISE, servo_id, data)

    def get_servo_cww(self, servo_id):
        """Reads the counterclockwise insensitive area of the encoder of the specified servo motor

        Args:
            servo_id (int): 0 - 254
        """
        return self._mesg(ProtocolCode.GET_SERVO_MOTOR_COUNTER_CLOCKWISE, servo_id, has_reply=True)

    def set_servo_system_data(self, servo_id, addr, data, mode):
        """Set the system parameters for the specified servo motor

        Args:
            servo_id (int): 0 - 254:
            addr (int):
            data (int): 0 - 4096
            mode (int): 1 - data 1byte. 2 - data 2byte
        """
        if mode == 1:
            self._mesg(ProtocolCode.SET_SERVO_MOTOR_CONFIG, servo_id, addr, data, mode)
        elif mode == 2:
            self._mesg(ProtocolCode.SET_SERVO_MOTOR_CONFIG, servo_id, addr, [data], mode)

    def get_servo_system_data(self, servo_id, addr, mode):
        """Read the system parameters of the specified servo motor

        Args:
            servo_id (int): 0 - 254
            addr (int):
            mode (int): 1 - data 1byte. 2 - data 2byte
        """
        return self._mesg(ProtocolCode.GET_SERVO_MOTOR_CONFIG, servo_id, addr, mode, has_reply=True)

    def set_master_out_io_state(self, io_number, status=1):
        """Set the host I/O pin status

        Args:
            io_number: 1 - 2
            status: 0/1; 0: low; 1: high. default: 1

        """
        self._mesg(ProtocolCode.SET_MASTER_PIN_STATUS, io_number, status)

    def get_master_in_io_state(self, io_number):
        """get the host I/O pin status

        Args:
            io_number (int): 1 - 2

        Returns:
                0 or 1. 1: high 0: low
        """
        return self._mesg(ProtocolCode.GET_MASTER_PIN_STATUS, io_number, has_reply=True)

    def set_tool_out_io_state(self, io_number, status=1):
        """Set the Atom pin status

        Args:
            io_number (int): 1 - 2
            status: 0 or 1; 0: low; 1: high. default: 1

        """
        self._mesg(ProtocolCode.SET_ATOM_PIN_STATUS, io_number, status)

    def get_tool_in_io_state(self, io_number):
        """Get the Atom pin status

        Args:
            io_number (int): pin number

        Returns:
            0 or 1. 1: high 0: low

        """
        return self._mesg(ProtocolCode.GET_ATOM_PIN_STATUS, io_number, has_reply=True)

    def set_tool_led_color(self, r, g, b):
        """Set the Atom LED color

        Args:
            r: 0-255
            g: 0-255
            b: 0-255

        """
        self._mesg(ProtocolCode.GET_ATOM_LED_COLOR, r, g, b)

    def is_tool_btn_clicked(self):
        """read the atom press status

        Returns:
            int: 0 or 1. 1: press

        """
        return self._mesg(ProtocolCode.GET_ATOM_PRESS_STATUS, has_reply=True)



# coding=utf-8

import logging
import logging.handlers


def setup_logging(debug=False):
    root_logger = logging.getLogger()

    debug_fomatter = logging.Formatter(
        fmt="%(asctime)s.%(msecs)03d %(levelname).4s [%(name)s] %(message)s",
        datefmt="%H:%M:%S",
    )
    logger_handle = logging.StreamHandler()
    logger_handle.setFormatter(debug_fomatter)
    if debug:
        logger_handle.setLevel(logging.DEBUG)
        # 100M日志
        save = logging.handlers.RotatingFileHandler(
        "python_debug.log", maxBytes=100*1024*1024, backupCount=1)
        save.setFormatter(debug_fomatter)
        root_logger.addHandler(save)
    else:
        logger_handle.setLevel(logging.WARNING)

    root_logger.addHandler(logger_handle)
    root_logger.setLevel(0)



# coding=utf-8

from __future__ import print_function

try:
    from serial.tools.list_ports import comports
except ImportError:
    print("Not found module `pyserial`, please install.")


def get_port_list():
    # type:() -> list
    """Get all the serial port string.

    Return:
        serial port list(list)
    """
    return [p.device for p in comports()]


def detect_port_of_basic():
    """Detect M5 Basic.
    Returns the serial port string of the first detected M5 Basic.
    If it is not found, it returns `None`.
    """
    ports = [p.device for p in comports() if p.pid == 0xEA60]
    return ports[0] if len(ports) > 0 else None
